/*
 * Copyright (c) 2018 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Mon Aug 20 2018
 */
#include "acmccdaqft.h"

#include <fcntl.h>
#include <math.h>
#include <unistd.h>

#include "mccdaqft_c_types.h"
#include "codels.h"

/* default force/torque sensor calibration */
static const struct {
  double calibration[6 * 6];
  double scale[6];
  or_t3d_pos poffset;
  or_t3d_att qoffset;
} daqft_ATI_FT13244 = {

  .calibration = {
    0.79596,   0.38534,   2.24097, -28.93308,  -1.72458,  32.52832,
    -3.31933, 39.97694,   1.37874, -16.72431,   0.76613, -19.19259,
    19.32599,  1.27589,  19.65418,   0.64218,  18.62043,   1.70901,
    0.09429,   0.15639, -33.04258,  -0.97552,  32.80478,   2.73001,
    37.63401,  2.41535, -19.56219,  -0.56291, -19.32668,  -1.78456,
    1.93227, -18.83459,   1.64382, -16.86959,   0.22141, -19.32653
  },

  .scale = {
    1.4229753385797,
    1.4229753385797,
    0.568322187462829,
    60.448865969817,
    60.448865969817,
    55.3363491382083
  },

  .poffset = { .x = 0, .y = 0, .z = 0.006858 },
  .qoffset = { .qw = 1., .qx = 0., .qy = 0, .qz = 0. }
};


/* --- Activity connect ------------------------------------------------- */

/** Validation codel daqft_check_rate of activity connect.
 *
 * Returns genom_ok.
 * Throws mccdaqft_e_nodev, mccdaqft_e_baddev, mccdaqft_e_inval.
 */
genom_event
daqft_check_rate(uint32_t rate, const genom_context self)
{
  return rate < 1 ? mccdaqft_e_inval(self) : genom_ok;
}


/* --- Function set_frame_offset ---------------------------------------- */

/** Codel daqft_set_offset of function set_frame_offset.
 *
 * Returns genom_ok.
 */
genom_event
daqft_set_offset(double x, double y, double z, double yaw,
                 double pitch, double roll,
                 mccdaqft_ids_calibration_s *calibration,
                 const genom_context self)
{
  (void)self; /* -Wunused-parameter */

  double cr = cos(roll/2.), cp = cos(pitch/2.), cy = cos(yaw/2.);
  double sr = sin(roll/2.), sp = sin(pitch/2.), sy = sin(yaw/2.);

  calibration->p = (or_t3d_pos){
    .x = x, .y = y, .z = z
  };
  calibration->q = (or_t3d_att){
    .qw = cr * cp * cy + sr * sp * sy,
    .qx = sr * cp * cy - cr * sp * sy,
    .qy = cr * sp * cy + sr * cp * sy,
    .qz = cr * cp * sy - sr * sp * cy
  };

  daqft_get_calibration(&daqft_ATI_FT13244.poffset,
                        &daqft_ATI_FT13244.qoffset,
                        daqft_ATI_FT13244.calibration,
                        daqft_ATI_FT13244.scale,
                        &calibration->p, &calibration->q,
                        calibration->G);

  return genom_ok;
}


/* --- Function log ----------------------------------------------------- */

/** Codel daqft_log_start of function log.
 *
 * Returns genom_ok.
 * Throws mccdaqft_e_sys.
 */
genom_event
daqft_log_start(const char path[64], uint32_t decimation,
                mccdaqft_log_s **log, const genom_context self)
{
  int fd;

  fd = open(path, O_WRONLY|O_APPEND|O_CREAT|O_TRUNC, 0666);
  if (fd < 0) return mccdaqft_e_sys_error(path, self);

  if (write(fd, mccdaqft_log_header "\n", sizeof(mccdaqft_log_header)) < 0)
    return mccdaqft_e_sys_error(path, self);

  if ((*log)->req.aio_fildes >= 0) {
    close((*log)->req.aio_fildes);

    if ((*log)->pending)
      while (aio_error(&(*log)->req) == EINPROGRESS)
        /* empty body */;
  }
  (*log)->req.aio_fildes = fd;
  (*log)->pending = false;
  (*log)->skipped = false;
  (*log)->decimation = decimation < 1 ? 1 : decimation;
  (*log)->missed = 0;
  (*log)->total = 0;

  return genom_ok;
}


/* --- Function log_stop ------------------------------------------------ */

/** Codel daqft_log_stop of function log_stop.
 *
 * Returns genom_ok.
 */
genom_event
daqft_log_stop(mccdaqft_log_s **log, const genom_context self)
{
  (void)self; /* -Wunused-parameter */

  if (*log && (*log)->req.aio_fildes >= 0)
    close((*log)->req.aio_fildes);
  (*log)->req.aio_fildes = -1;

  return genom_ok;
}


/* --- Function log_info ------------------------------------------------ */

/** Codel daqft_log_info of function log_info.
 *
 * Returns genom_ok.
 */
genom_event
daqft_log_info(const mccdaqft_log_s *log, uint32_t *miss,
               uint32_t *total, const genom_context self)
{
  (void)self; /* -Wunused-parameter */

  *miss = *total = 0;
  if (log) {
    *miss = log->missed;
    *total = log->total;
  }
  return genom_ok;
}
